-- Hazara Khan's implementation of the Suicide Kings loot distribution
--  system. See http://www.wowwiki.com/Suicide_Kings for a description
--  of the system, and <Seraphim> of Medivh's loot rules for the basis
--  for this implementation.
-- This is intended to be a module for HazLoot, a master looter aid, but
--  written in such a way that it's testable outside of WoW.

-- TODO: With HSK:Capitalize, consider a HSK.DisplayName table

debug = false

HSK = {}
if not HazLootList then -- not in WoW, because Ace creates HazLootList
  HSK.standalone = true
end

-- DATA STRUCTURES
HSK.MasterList   = {} -- The master list    {[index] = memberName}
HSK.InvertedList = {} -- Inverted master list {[memberName]=index}
HSK.Alts         = {} -- Alts of members' mains in the master list
                      --  {[memberName]={alt1,alt2,...altN,}}
HSK.Mains        = {} -- Mains of members' alts (inverse of Alts)
HSK.Present      = {} -- A sub-list of the master, of present raiders' mains
HSK.InvertedP    = {} -- Inverted present list {[main]=presentKey} (mains are lowercase)
HSK.PresentCount = 0  -- Present is not 1,2,..n-indexed, have to count size
HSK.Roles        = {} -- Raiders' declared roles {[name]=role}, includes alts.
HSK.RoleSet      = {} -- Set of permissible roles {[role]=count} where count is of Mains

-- OPTIONS
HSK.DropOnMainChange = true -- This is Seraphim's default
HSK.DropOnRoleChange = true -- This is Seraphim's default

--- STRATEGY INTERFACE ---
function HSK:GetName()
  return "Suicide Kings (Seraphim)"
end

--- Return the winner from an array of members of the Present list.
function HSK:GetWinner(array)
  if not array then return nil end
  if not HSK.Present then return nil end
  if 0 == HSK.PresentCount then return nil end
  local winner = #HSK.MasterList + 2
  local j = nil
  for i,candidate in ipairs(array) do
    candidate = HSK:Capitalize(candidate)
    if HSK:IsPresent(candidate) then
      j = HSK:GetMasterPosition(candidate)
      winner = (j < winner) and j or winner
    end
  end
  return HSK.MasterList[winner]
end

--- Does a member fit a role? Some strategies restrict bids and rolls by role
-- @param member A raid member
-- @param role A role
-- @returns true if both member and role are valid and member is of role
function HSK:IsMemberInRole(member, role)
  if not member or not role then return false end
  return role == HSK:GetRole(member)
end

--- Sets the permissible roles.
-- All members must have a raid role, and each role must belong to the set of known
-- (or permissible) roles.
-- @param roles A string array of raid role names.
-- @return The number of roles added.
function HSK:SetRoles(roles)
  if not roles then return nil end
  local rolecount = 0 -- # operator FAILs on non-array tables. Crappy fucking language.
  for i,role in ipairs(roles) do
    HSK.RoleSet[role] = 0
    rolecount = rolecount + 1
  end
  return rolecount
end

--- Adds a new role
-- @param role The role to add
-- @returns true if role was added, or already exists, nil if role is invalid
function HSK:AddRole(role)
  if not role or "string" ~= type(role) then return nil end
  if HSK.RoleSet[role] then return true end
  HSK.RoleSet[role] = 0
  return nil
end

--- How many toons (mains and alts) have this role
function HSK:GetRoleCount(role)
  if not role then return 0 end
  if not HSK.RoleSet[role] then return 0 end
  return HSK.RoleSet[role]
end

--- Add a new member to the Suicide Kings list at an (optional) position
-- Attempts to append a new member to the bottom of the list, or at the
--  position provided. If position would leave a gap at the bottom of the
--  list, then the member is added at the bottom, leaving no gaps.
-- @param member   (Required) The main character name of a raider in the SK List
-- @param role     (Required) The main's declared raiding role.
-- @param position (Optional) The position in the list of the new member. All existing
--                 members at, or below, this position will be shuffled down one slot.
--                 Position must be > 0 If position is omitted, the new member starts
--                 at the bottom of the list.
-- @return The new position of the added member, nil if addition failed, the member or
--          role are undefined, or the role is not known.
function HSK:AddMember(member, role, position)
  if not member or not role then return nil end
  member = HSK:Capitalize(member)
  if not HSK.RoleSet[role] then return nil end
  if HSK.InvertedList[member] then
    return HSK.InvertedList[member]
  end
  local listsize = (#HSK.MasterList)
  if nil == position or listsize < position then -- append
    table.insert(HSK.MasterList, member)
    position = (#HSK.MasterList)
    HSK.InvertedList[member] = position
  elseif 0 < position and position <= listsize then
    for i = position,listsize do
      HSK.InvertedList[HSK.MasterList[i]] = i+1
    end
    table.insert(HSK.MasterList, position, member)
    HSK.InvertedList[member] = position
  else
    return nil
  end
  HSK.Roles[member] = role
  HSK.RoleSet[role] = HSK.RoleSet[role] + 1
  return HSK.InvertedList[member]
end

--- Is the toon a main or alt in this SK List?
-- @returns true if so, false if not, nil if bad input
function HSK:IsMember(member)
  if nil == member then return nil end
  if HSK.InvertedList[member] or HSK.Mains[member] then
    return true
  end
  return nil
end

--- Returns the raid role of a member (or alt)
-- @param member (Required) The character's name. May be main or alt.
-- @returns the declared raid role, nil if the member is not known.
function HSK:GetRole(member)
  if nil == member then return nil end
  if debug then self:Debug("[GetRole("..member..")]") end
  local found = HSK:GetMasterPosition(member)
  if debug then self:Debug("[GetRole("..member..")] found = "..found) end
  if found then
    return HSK.Roles[member]
  end
  return nil
end

--- Clears the master, present and alt lists, the role lists, and role counts
-- @returns true on success
function HSK:Clear()
  HSK.MasterList   = {}
  HSK.InvertedList = {}
  HSK.Alts         = {}
  HSK.Mains        = {}
  HSK.Present      = {}
  HSK.InvertedP    = {}
  HSK.PresentCount = 0 
  HSK.Roles        = {}
  HSK.RoleSet      = {}
  return true
end

--- Replace the master list with an ordered list of raiders.
-- @param members An array of the members of an SK list, in suicide order, with their roles:
--                {"member1","role1","member2","role2",..."memberN","roleN"}
-- @param alts (Optional) A table of [member]={alt1,alt2,...altN,} entries. Keys in the alts
--                        table must exist in the members array parameter. May be nil or {}
-- @return The number of members added if the list was created correctly, nil otherwise
function HSK:SetMasterList(members, alts)
  if not "table"==type(members) then return false end
  if alts and not "table" == type (alts) then return false end
  HSK:Clear()
  for i,main in ipairs(members) do
    if 1 == i%2 then
      main = HSK:Capitalize(main)
      HSK:AddRole(members[i+1])
      HSK:AddMember(main, members[i+1])
    end
  end
  if alts then
    for member, altList in pairs(alts) do
      for i,alt in ipairs(altList) do
        if 1 == i%2 then
          alt = HSK:Capitalize(alt)
          HSK:AddRole(altList[i+1])
          HSK:AddAlt(member, alt,altList[i+1])
        end
      end
    end
  end
  return HSK:GetMasterListSize()
end

--- Remove a member from the Suicide Kings list
-- Permanently removes a member from the SK List - for example when they leave
-- the guild, stop raiding, are an utter douchebag, whatever. Adjusts the list
-- so that the list has no gaps.
-- @param member The name of a member to remove, or their position in the list.
-- @param confirm true|false If true, remove the main, and all alts corresponding
--                to member. Otherwise, return the name of the main that would
--                be nuked.
-- @return true if the member was successfully removed, the name of the main that
--         would be removed if confirm is not true, nil if not found
function HSK:RemoveMember(member, confirm)
  if not member then return false end
  local position
  local membername
  if "string" == type(member) then
    position = HSK:GetMasterPosition(member)
  elseif "number" == type(member) then
    position = member
  end
  if position >= 0 and position < (#HSK.MasterList) then
    membername = HSK:GetMemberAt(position) -- may be an alt
    if not confirm then return membername end -- must have confirmation
    if HSK.Alts[membername] then
      for i,alt in ipairs(HSK.Alts[membername]) do
        HSK:RemoveAlt(alt)
      end
    end
    HSK.Alts[membername] = nil
    HSK.RoleSet[HSK.Roles[membername]] = HSK.RoleSet[HSK.Roles[membername]] - 1
    HSK.Roles[membername] = nil
    HSK.InvertedList[membername] = nil
    local listsize = (#HSK.MasterList)
    for i = position+1,listsize do
      HSK.MasterList[i-1] = HSK.MasterList[i]
      HSK.InvertedList[HSK.MasterList[i]] = i-1
    end
    table.remove(HSK.MasterList, listsize)
    return true
  end
  return false
end

--- Add an alt to the SK List, associated with its corresponding main.
--  @param main The name of the correpsonding main
--  @param alt  The name of the alt in question
--  @param role (Optional) The role of the alt in question, defaults to main's role
--  @return The position of the main in the SK List, nil if the main was
--          not found, or the alt is already on another main, or an invalid
--          role was specified
function HSK:AddAlt(main, alt, role)
  if not main or not alt then return nil end
  if not role then role = HSK.Roles[main] end
  if not HSK.RoleSet[role] then return nil end
  local mainPos = HSK:GetMasterPosition(main)
  local altPos  = HSK:GetMasterPosition(alt)
  if altPos and mainPos == altPos then return altPos -- Already exists
  elseif altPos then -- Alt exists on another main
    return nil
  end
  if mainPos then 
    if not HSK.Alts[main] then HSK.Alts[main] = {alt}
    else
      table.insert(HSK.Alts[main],alt)
    end
    HSK.Mains[alt] = main
    HSK.Roles[alt] = role
    HSK.RoleSet[role] = HSK.RoleSet[role] + 1
  end
  return mainPos
end

--- Remove an alt from the records, leaving the main unaffected
--  @param alt The name of the alt in question
--  @return true if the alt was removed, false if the alt was a
--          main, nil if the alt was not found
function HSK:RemoveAlt(alt)
  if HSK.InvertedList[alt] then return false end
  local main = HSK.Mains[alt]
  if not main then return nil end
  for i,anAlt in ipairs(HSK.Alts[main]) do
    if alt == anAlt then
      HSK.RoleSet[HSK.Roles[alt]] = HSK.RoleSet[HSK.Roles[alt]] - 1
      table.remove(HSK.Alts[main],i)
      HSK.Mains[alt] = nil
      return true
    end
  end
  return nil
end

--- Promote an alt to main, and move the corresponding main to the alt list.
-- @param alt The name of an alt to promote
-- @return The position of the new main in the SK List. In Seraphim, changing main
--         means a drop to the bottom of the list. This is configurable. Returns nil
--         if no alt is specified. No-op if given the name of a main.
-- @see SetDropOnMainChange
function HSK:PromoteAlt(alt)
  if nil == alt then return nil end
  local oldMain = HSK:GetMain(alt)
  if alt == oldMain then return HSK.InvertedList[oldMain] end -- Already a main
  local newMain = alt
  -- Switch them in the inverted list
  HSK.InvertedList[newMain]  = HSK.InvertedList[oldMain]
  HSK.InvertedList[oldMain] = nil
  -- Switch them in the Mains table
  HSK.Mains[oldMain] = newMain
  HSK.Mains[newMain] = nil
  -- Point newMain's alt table at oldMain's alt table
  HSK.Alts[newMain] = HSK.Alts[oldMain]
  -- Remove reference to oldMain's alt table
  HSK.Alts[oldMain] = nil
  -- For each member, set its main to newMain, noting which was newMain
  local newMainOldIndex = 0
  for i,anAlt in ipairs(HSK.Alts[newMain]) do
    if newMain == anAlt
      then newMainOldIndex = i
    else
      HSK.Mains[anAlt] = newMain
    end
  end
  -- Remove newMain from its own alt list, and add oldMain
  table.remove(HSK.Alts[newMain],newMainOldIndex)
  table.insert(HSK.Alts[newMain],oldMain)

  -- If we drop on main or role change, drop
  if HSK.DropOnMainChange then
    return HSK:Drop(alt)
  elseif HSK.DropOnRoleChange and HSK.Roles[oldMain] ~= HSK.Roles[newMain] then
    return HSK:Drop(alt)
  else
    return HSK.InvertedList[alt]
  end
  return nil
end

--- Changes the raid role of a main or alt
-- Depending on how the SK List is configured, changing a main's role may drop the
--  main to the bottom of the SK List, by default. Nothing happens if the new role
--  is the same as the old role.
-- @param member The toon changing role
-- @param role   The new role
-- @returns The new role, or nil if member was invalid or not found, or the role was unknown
-- @see SetDropOnRoleChange
function HSK:ChangeRole(member,role)
  if not member then return nil end
  local main = HSK:GetMain(member)
  local oldRole = HSK.Roles[member]
  if role == oldRole then return oldRole -- no change
  else
    HSK.RoleSet[oldRole] = HSK.RoleSet[oldRole] - 1
    HSK.Roles[member]    = role
    HSK.RoleSet[role]    = HSK.RoleSet[role] + 1
  end
  if main == member and HSK.DropOnRoleChange then
    HSK:Drop(member)
  end
end

--- Returns an array of a member's alts.
-- @param main The main whose alts you want
-- @returns A cloned array of the member's alts, empty if there are none,
--          nil if main is invalid, not found or an alt
function HSK:GetAlts(main)
  if not main then return nil end
  if not HSK.InvertedList[main] then return nil end
  local alts = {}
  if HSK.Alts[main] then 
    for i,alt in ipairs(HSK.Alts[main]) do
      table.insert(alts, alt)
    end
  end
  return alts
end

--- Returns the main of a member (alt or main)
function HSK:GetMain(member)
  if not member then return nil end
  if HSK.InvertedList[member] then return member end
  return HSK.Mains[member]
end

--- Return the number of known toons (Main + Alts) of a member
-- @param member The name of an alt or main.
-- @returns Number of known toons (1 + #Alts) for member, nil if
--          member is nil, 0 if member is not known
function HSK:GetToonCount(member)
  if not member then return nil end -- bad input
  local mainPos = HSK:GetMasterPosition(member)
  if not mainPos then return 0 end  -- not found
  local alts = HSK.Alts[HSK.MasterList[mainPos]]
  if not alts then return 1 end
  return 1 + #alts
end

--- Set an existing member to "present", and thus eligible for suicides
-- @param member The name of a member to mark as present, or an alt
-- @return The member's position in the master list, nil if not found
function HSK:MarkAsPresent(member)
  if not member then return nil end
  member = HSK:Capitalize(member)
  local main = HSK:GetMain(member)
  if nil == main then return nil end
  if HSK.InvertedP[main] then return HSK.InvertedP[main] end
  HSK.InvertedP[main] = HSK.InvertedList[main]
  HSK.Present[HSK.InvertedP[main]] = main
  HSK.PresentCount = HSK.PresentCount + 1
  return HSK.InvertedP[main]
end

--- Is the member present in the raid?
-- @param member The name of a main or alt
-- @returns true if the member is present, the name of an alt's main if member is
--          an alt or false if the member is not present
function HSK:IsPresent(member)
  if not member then return nil end
  member = HSK:Capitalize(member)
  local main = HSK:GetMain(member)
  if not main then return nil end
  if main ~= member then return main end
  return (nil ~= HSK.InvertedP[main] and true or nil)
end

--- Return a member's current position in the master SK List
--  @param position Integer position in the list.
--  @return The name of the member in the position provided, or nil if position
--          is larger than the size of the list.
function HSK:GetMemberAt(position)
  return HSK.MasterList[position]
end

--- Return the current position in the master list of a raid member
--  @param member A main in the master list, or one of their alts.
--  @return The position in the master list of a raid member, or the position in
--          the master list of a main of one of that member's alts. Returns nil
--          if member is not in the master list, nor an alt.
function HSK:GetMasterPosition(member)
  if nil == member then return nil end
  local main = HSK.Mains[member]
  if debug then self:Debug("GetMasterPos: HSK.InvertedList["..member.."]: "..(HSK.InvertedList[member] or "nil")) end
  if debug then self:Debug("GetMasterPos: HSK.InvertedList["..(main or "nil").."]: is next") end
  return HSK.InvertedList[member] or HSK.InvertedList[main]
end

--- Suicide an active member, and adjust the master SK List.
--  @param member A main that has been marked as present, or one of their alts.
--  @return The new position in the list, nil if they were not found, or there's no
--          present list.
function HSK:Suicide(member)
  if not member then return nil end
  member = HSK:Capitalize(member)
  if not HSK.Present or 0 == HSK.PresentCount then return nil end
  return HSK:Drop(HSK:GetMain(member), HSK.Present)
end

--- Drops an active member to the bottom of the Master SK List
--  @param member The main or alt of a main that's being dropped
--  @param list   The list in which to drop member (Master if nil)
--  @return their new position in the list, nil if they were not found.
function HSK:Drop(member, list)
  if not list then list = HSK.MasterList end
  local pos = HSK:GetMasterPosition(member)
  if (list == HSK.MasterList) then
    for i,main in ipairs(list) do -- This is a special case of the loop below,
      if i >= pos then            --  and it would be elegant to merge them
        list[i] = list[i+1] or member
      end
    end
  else -- This is a present list
    local index = {}
    for i,member in pairs(HSK.Present) do table.insert(index,i) end
    table.sort(index)
    for i,val in ipairs(index) do
      if i == #index then
        HSK.MasterList[val] = HSK.Present[pos]
      elseif val >= pos then
        HSK.MasterList[val] = HSK.Present[index[i+1]]
      end
    end
  end
  -- rebuild the inverted list
  HSK.InvertedList = {}
  for i,main in ipairs(HSK.MasterList) do
    HSK.InvertedList[main] = i
  end
  -- rebuild the present list, if there was one
  if (HSK.PresentCount > 0)  then
    local tmpPresent = {}
    for i,present in pairs(HSK.Present) do
      table.insert(tmpPresent, present)
    end
    HSK.Present = {}
    HSK.InvertedP = {}
    HSK.PresentCount = 0
    for i,present in pairs(tmpPresent) do
      HSK:MarkAsPresent(present)
    end
  end
  return HSK.InvertedList[member]
end

--- How many people are in the master list?
--  @return The size of the master list, 0 if empty.
function HSK:GetMasterListSize()
  return (#HSK.MasterList)
end

--- How many people are in the "present" list?
--  @return The size of the present list, 0 if empty.
function HSK:GetPresentCount()
  return (HSK.PresentCount)
end

--- Returns (cloned) arrays of the master and alts list, with their roles.
--  The lists returned can be persisted externally for subsequent restoration with
--   SetMasterList()
--  @return Cloned copies of the master SK table and the corresponding Alts with roles
--  @see SetMasterList
function HSK:GetMasterList()
-- Test.BulkMains = {"Main1","R1","Main2","R2","Main3","R3","Main4","R1","Main5","R2","Main6","R3","Main7","R1","Main8","R2","Main9","R3","Main10","R1",}
-- Test.BulkAlts  = {["Main1"]={"Main1Alt1","R1","Main1Alt2","R2",},["Main3"]={"Main3Alt1","R1","Main3Alt2","R2","Main3Alt3","R3",},}
  local masterList = {}
  local altList    = {}
  if HSK.MasterList then 
    for i,main in ipairs(HSK.MasterList) do
      table.insert(masterList, main)
      table.insert(masterList, HSK.Roles[main])
      if HSK.Alts[main] then
        altList[main] = {}
        for j,alt in ipairs(HSK.Alts[main]) do
          table.insert(altList[main],alt)
          table.insert(altList[main],HSK.Roles[alt])
        end
      end
    end
  end
  return masterList, altList
end

--- Return a (cloned) array of the list of members marked "present" for a raid.
--  The list returned is a copy. It should not therefore be manipulated with any
--  expectation of changes being reflected to the actual list, but is
--  appropriate for display, etc.
--  @return A cloned copy of the list of "present" raiders, in SK order, which
--          may be empty.
function HSK:GetPresentList()
  local list  = {}
  local index = {}
  if HSK.Present then 
    for i,main in pairs(HSK.Present) do
      table.insert(index,i)
      list[i] = main
    end
  end
  table.sort(index)
  return index, list
end

--- Export the master list to some format
-- By default, returns a string of CSV, for each main: "Main","Role","Alt1","Alt1Role","Alt2",..
-- TODO: write a corresponding import
-- @params format Currently ignored (defaults to CSV)
function HSK:ExportMasterList(format)
  local export = ""
  for i,main in ipairs(HSK.MasterList) do
    export = export.."\""..main.."\",\""..HSK:GetRole(main).."\""
    if HSK.Alts[main] then
      for j,alt in ipairs(HSK.Alts[main]) do
        export = export..",\""..alt.."\",\""..HSK:GetRole(alt).."\""
      end
    end
    export = export.."\n"
  end
  return export
end

--- Export the present list to some format
-- By default, returns a string of CSV, for each main: "Main","Role","Alt1","Alt1Role","Alt2",..
-- @params format Currently ignored (defaults to CSV)
function HSK:ExportPresentList(format)
  local export = ""
  for i,main in pairs(HSK.Present) do
    export = export.."\""..i.."\",\""..main.."\",\""..HSK:GetRole(main).."\""
    --[[
    if HSK.Alts[main] then
      for j,alt in ipairs(HSK.Alts[main]) do
        export = export..",\""..alt.."\",\""..HSK:GetRole(alt).."\""
      end
    end
    ]]--
    export = export.."\n"
  end
  return export
end

--- Bootstrap the SKList from Bootstrap.lua
--  To start using HazLoot's SK implementation, you can either edit Bootstrap.lua or use the
--  default, empty list and the (as yet unimplemented) UI to add people manually
-- @params tableSet Pair of tables in Bootstrap.lua to use to bootstrap
-- @returns true if succesfully bootstrapped, false otherwise
function HSK:Bootstrap(tableSet)
  if not tableSet then return false end
  local masterList, altList = Bootstrap:GetBootTables(tableSet)
  HSK:SetMasterList(masterList, altList)
  return false
end

--- Should the module drop a member to the bottom of the list when changing mains
function HSK:GetDropOnMainChange()
  return HSK.DropOnMainChange
end
--- Set whether the module should drop a member to the bottom of the list when changing mains
function HSK:SetDropOnMainChange(boolean)
  if nil == boolean then return nil end
  HSK.DropOnMainChange = boolean
end

--- Should the module drop a member to the bottom of the list when changing roles
function HSK:GetDropOnRoleChange()
  return HSK.DropOnRoleChange
end
--- Set whether the module should drop a member to the bottom of the list when changing roles
function HSK:SetDropOnRoleChange(boolean)
  if nil == boolean then return nil end
  HSK.DropOnRoleChange = boolean
end

--- Player names have initial capitals, and all lowercase thereafter.
function HSK:Capitalize(unit)
  if nil == unit then return nil end
  if "string" ~= type(unit) then return unit end
  unit = string.upper(string.sub(unit,1,1))..string.lower(string.sub(unit,2))
  return unit
end

function HSK:Print(message)
  if HazLoot then
    HazLoot:Print(message)
  else
    print("[HSK] "..message)
  end
end

function HSK:Debug(message)
  if HazLoot then
    HazLoot:Debug(message)
  else
    print("[HSK][DEBUG] "..message)
  end
end
